import numpy as np
import torch


def area_of(left_top, right_bottom) -> torch.Tensor:
    """Compute the areas of rectangles given two corners.

    Args:
        left_top (N, 2): left top corner.
        right_bottom (N, 2): right bottom corner.

    Returns:
        area (N): return the area.
    """
    hw = torch.clamp(right_bottom - left_top, min=0.0)
    return hw[..., 0] * hw[..., 1]


def iou_of(boxes0, boxes1, eps=1e-5):
    """Return intersection-over-union (Jaccard index) of boxes.

    Args:
        boxes0 (N, 4): ground truth boxes.
        boxes1 (N or 1, 4): predicted boxes.
        eps: a small number to avoid 0 as denominator.
    Returns:
        iou (N): IoU values.
    """
    overlap_left_top = torch.max(boxes0[..., :2], boxes1[..., :2])
    overlap_right_bottom = torch.min(boxes0[..., 2:], boxes1[..., 2:])

    overlap_area = area_of(overlap_left_top, overlap_right_bottom)
    area0 = area_of(boxes0[..., :2], boxes0[..., 2:])
    area1 = area_of(boxes1[..., :2], boxes1[..., 2:])
    return overlap_area / (area0 + area1 - overlap_area + eps)


def compute_average_precision(precision, recall):
    """
    It computes average precision based on the definition of Pascal Competition. It computes the under curve area
    of precision and recall. Recall follows the normal definition. Precision is a variant.
    pascal_precision[i] = typical_precision[i:].max()
    """
    # identical but faster version of new_precision[i] = old_precision[i:].max()
    precision = np.concatenate([[0.0], precision, [0.0]])
    for i in range(len(precision) - 1, 0, -1):
        precision[i - 1] = np.maximum(precision[i - 1], precision[i])

    # find the index where the value changes
    recall = np.concatenate([[0.0], recall, [1.0]])
    changing_points = np.where(recall[1:] != recall[:-1])[0]

    # compute under curve area
    areas = (recall[changing_points + 1] - recall[changing_points]) * precision[changing_points + 1]
    return areas.sum()


def compute_voc2007_average_precision(precision, recall):
    ap = 0.
    for t in np.arange(0., 1.1, 0.1):
        if np.sum(recall >= t) == 0:
            p = 0
        else:
            p = np.max(precision[recall >= t])
        ap = ap + p / 11.
    return ap


def compute_average_precision_per_class(num_true_cases, gt_boxes, difficult_cases,
                                        prediction_file, iou_threshold, use_2007_metric,
                                        name="", plot=True):
    with open(prediction_file) as f:
        image_ids = []
        boxes = []
        scores = []
        for line in f:
            t = line.rstrip().split(" ")
            image_ids.append(t[0])
            scores.append(float(t[1]))
            box = torch.tensor([float(v) for v in t[2:]]).unsqueeze(0)
            box -= 1.0  # convert to python format where indexes start from 0
            boxes.append(box)
        scores = np.array(scores)
        sorted_indexes = np.argsort(-scores)
        boxes = [boxes[i] for i in sorted_indexes]
        image_ids = [image_ids[i] for i in sorted_indexes]
        true_positive = np.zeros(len(image_ids))
        false_positive = np.zeros(len(image_ids))
        matched = set()
        for i, image_id in enumerate(image_ids):
            box = boxes[i]
            if image_id not in gt_boxes:
                false_positive[i] = 1
                continue

            gt_box = gt_boxes[image_id]
            ious = iou_of(box, gt_box)
            max_iou = torch.max(ious).item()
            max_arg = torch.argmax(ious).item()
            if max_iou > iou_threshold:
                if difficult_cases[image_id][max_arg] == 0:
                    if (image_id, max_arg) not in matched:
                        true_positive[i] = 1
                        matched.add((image_id, max_arg))
                    else:
                        false_positive[i] = 1
            else:
                false_positive[i] = 1

    true_positive = true_positive.cumsum()
    false_positive = false_positive.cumsum()
    precision = true_positive / (true_positive + false_positive)
    recall = true_positive / num_true_cases
    if use_2007_metric:
        ap = compute_voc2007_average_precision(precision, recall)
    else:
        ap = compute_average_precision(precision, recall)
    if plot:
        sorted_scores = scores[sorted_indexes]
        plot_map_recall_precision(ap, recall, precision, sorted_scores, name)
    return ap


def plot_map_recall_precision(ap, recall, precision, sorted_scores, name):
    """
    sorted_scores = scores[sorted_indexes]
    :param ap:
    :param recall:
    :param precision:
    :param sorted_scores:
    :param name:
    :return:
    """
    import matplotlib.pyplot as plt
    # 逆序排序
    recall = [0] + recall.tolist() + [1]
    precision = [1] + precision.tolist() + [0]
    sorted_scores = [1] + sorted_scores.tolist() + [0]
    plt.figure()
    plt.plot(recall, precision, label='Precision')
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    ap_str = "{0:.2f}%".format(ap * 100)
    # ap_str = "{0:.4f}%".format(ap * 100)
    plt.title('Precision x Recall curve \nClass: %s, AP: %s' % (name, ap_str))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.figure()
    plt.plot(sorted_scores, recall, label='Recall')
    plt.xlabel('Scores')
    plt.ylabel('Recall')
    plt.title('Recall-Scores\nClass:{}'.format(name))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.figure()
    plt.plot(sorted_scores, precision, label='Precision')
    plt.xlabel('Scores')
    plt.ylabel('Precision')
    plt.title('Precision-Scores\nClass:{}'.format(name))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.show()
    # plt.waitforbuttonpress()
    plt.pause(0.05)



# -*-coding: utf-8 -*-
"""
    @Project: python-learning-notes
    @File   : iou.py
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-07-10 10:14:56
"""

import numpy as np


def cal_iou(box1, box2):
    """
    :param box1: = [xmin1, ymin1, xmax1, ymax1]
    :param box2: = [xmin2, ymin2, xmax2, ymax2]
    :return:
    """
    xmin1, ymin1, xmax1, ymax1 = box1
    xmin2, ymin2, xmax2, ymax2 = box2
    # 计算每个矩形的面积
    s1 = (xmax1 - xmin1) * (ymax1 - ymin1)  # C的面积
    s2 = (xmax2 - xmin2) * (ymax2 - ymin2)  # G的面积

    # 计算相交矩形
    xmin = max(xmin1, xmin2)
    ymin = max(ymin1, ymin2)
    xmax = min(xmax1, xmax2)
    ymax = min(ymax1, ymax2)

    w = max(0, xmax - xmin)
    h = max(0, ymax - ymin)
    area = w * h  # C∩G的面积
    iou = area / (s1 + s2 - area)
    return iou


def cal_iou_list(box1, box2_list):
    iou_list = []
    for box2 in box2_list:
        iou = cal_iou(box1, box2)
        iou_list.append(iou)
    return iou_list


def get_iou_mat(box1_list, box2_list):
    '''
    获得IOU矩阵:
    :param box1_list: len1
    :param box2_list: len2
    :return:iou_mat = <np.ndarray>: (len1, len2)
    '''
    iou_mat = []
    for box1 in box1_list:
        iou_list = cal_iou_list(box1, box2_list)
        # iou_list = cal_bbox_distance_list(box1, box2_list)
        iou_mat.append(iou_list)
    iou_mat = np.asarray(iou_mat)
    return iou_mat


def get_max_iou_index(iou_mat, iou_threshold):
    max_index = np.argmax(iou_mat, axis=1)
    max_iou = np.max(iou_mat, axis=1)
    index = max_iou > iou_threshold
    max_iou = max_iou[index]
    max_index = max_index[index]
    return max_iou, max_index


def cal_distance(box1, box2):
    """
    :param box1: = [xmin1, ymin1, xmax1, ymax1]
    :param box2: = [xmin2, ymin2, xmax2, ymax2]
    :return:
    """
    xmin1, ymin1, xmax1, ymax1 = box1
    xmin2, ymin2, xmax2, ymax2 = box2
    center1 = np.asarray([xmin1 + xmax1, ymin1 + ymax1]) / 2.0
    center2 = np.asarray([xmin2 + xmax2, ymin2 + ymax2]) / 2.0
    d = np.sqrt(np.sum(np.power((center1 - center2), 2)))
    return d


def cal_bbox_distance_list(box1, box2_list):
    dist_list = []
    for box2 in box2_list:
        d = cal_distance(box1, box2)
        dist_list.append(d)
    return dist_list


if __name__ == '__main__':
    box1 = [661, 27, 679, 47]
    box2 = [661, 27, 679, 47]
    box2_list = []
    box2_list.append(box2)
    box2_list.append(box2)
    iou = cal_iou_list(box1, box2_list)
    print(iou)

if __name__ == '__main__':
    box1 = [661, 27, 679, 47]
    box2 = [661, 27, 679, 47]
    box2_list = []
    box2_list.append(box2)
    box2_list.append(box2)
    iou = cal_iou_list(box1, box2_list)
    print(iou)



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-23 14:45:13
    @Brief  :常见问题：
    (1) TRT多模型推理时，Pytorch模型可能会出现冲突，建议Pytorch模型不要使用DataParallel加载模型推理
    (2) 推理输入数据batch_size>1时，需要在转换ONNX模型时，设置dynamic=True
    (3) 多次`import pycuda.autoinit` 可能出现异常:Error Code 1: Cask (Cask convolution execution)
"""
import os
import cv2
import numpy as np
import pycuda.driver as cuda
import pycuda.autoinit
import tensorrt as trt
import traceback
from typing import Tuple, List

cuda.init()
device_context = cuda.Device(0).make_context()
TRT_LOGGER = trt.Logger(trt.Logger.ERROR)  # TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


class TRTEngine(object):
    """TensorRT引擎"""

    def __init__(self, model_file: str, input_shape: tuple, quant=1, calibrator=None):
        """
        https://www.jianshu.com/p/36ff0e224112
        :param model_file: 模型文件，可以是*.trt或者*.onnx文件
        :param input_shape:输入维度(B, C, H, W)
        :param quant: 0:不进行量化，1:进行半精度量化(FP16)，2:进行INT8量化(INT8)
        :param calibrator: 校准模块，INT8量化时需要
        """
        print("tensorrt:{}".format(trt.__version__))  # 8.4.1.5
        assert len(input_shape) == 4
        B, C, H, W = input_shape
        self.input_size = (W, H)
        self.batch_size = B
        self.calibrator = calibrator
        self.input_shape = input_shape
        # Load a serialized engine into memory
        self.engine = self.load_model(model_file=model_file, input_shape=self.input_shape, quant=quant)
        # Create context, this can be re-used  创建 执行环境
        self.context = self.engine.create_execution_context()
        # Profile 0 (first profile) is used by default  context可以设置多个profile， 这里选择第一个，也是默认的profile，其中规定了输入尺寸的变化区间
        self.context.active_optimization_profile = 0
        print("Active Optimization Profile: {}".format(self.context.active_optimization_profile))
        # These binding_idxs can change if either the context or the
        # active_optimization_profile are changed  获得输入输出变量名对应profile的idx
        self.input_binding_idxs, self.output_binding_idxs = self.get_binding_idxs(self.engine,
                                                                                  self.context.active_optimization_profile)
        print("\tinput_shape                    : {}".format(self.input_shape))
        # 获得输入变量的变量名
        self.input_names = [self.engine.get_binding_name(binding_idx) for binding_idx in self.input_binding_idxs]
        print("Input Metadata")
        print("\tNumber of Inputs: {}".format(len(self.input_binding_idxs)))
        print("\tInput names     : {}".format(self.input_names))
        print("\tInput Bindings for Profile {}: {}".format(self.context.active_optimization_profile,
                                                           self.input_binding_idxs))
        print("\tInput names: {}".format(self.input_names))
        # 获得输出变量的变量名
        print("Output Metadata")
        self.output_names = [self.engine.get_binding_name(binding_idx) for binding_idx in self.output_binding_idxs]
        print("\tNumber of Outputs: {}".format(len(self.output_binding_idxs)))
        print("\tOutput names     : {}".format(self.output_names))
        print("\tOutput Bindings for Profile {}: {}\n".format(self.context.active_optimization_profile,
                                                              self.output_binding_idxs))
        print("TRT model loaded successfully:{}".format(model_file.replace(".onnx", ".trt")))
        print('-----------' * 5, flush=True)

    def load_model(self, model_file: str, input_shape, quant=1):
        """
        :param model_file: ONNX或TRT模型文件
        :param input_shape: 模型输入维度
        :param quant: 0:不进行量化，1:进行半精度量化，2:进行INT8量化
        :return:
        """
        print('-----------' * 5)
        engine = None
        if model_file.endswith(".trt"):
            try:
                print('Load TRT model file         : {}'.format(model_file))
                engine = TRTEngine.load_engine(model_file)
                assert engine is not None, Exception('Load TRT model failed       : {}'.format(model_file))
                print('Load TRT Engine successfully: {}'.format(model_file))
            except Exception as e:
                traceback.print_exc()
                print('Load TRT model failed       : {}'.format(model_file), flush=True)
                model_file = model_file.replace(".trt", ".onnx")
        # TODO 如果TRT模型加载失败，则尝试从ONNX文件编译模型
        if model_file.endswith(".onnx") and os.path.exists(model_file):
            trt_file = model_file.replace(".onnx", ".trt")
            print('Try to build TRT Engine from ONNX file:{}'.format(model_file))
            engine = onnx2trt(model_file, trt_file=trt_file, input_shape=input_shape,
                              quant=quant, calibrator=self.calibrator)
            print("Completed parsing of ONNX file")
            print('Load TRT Engine successfully: {}'.format(trt_file))
        print('-----------' * 5, flush=True)
        return engine

    @staticmethod
    def load_engine(filename: str):
        # Load serialized engine file into memory 加载序列化的cuda引擎并进行反序列化
        with open(filename, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())

    def __call__(self, input_tensor):
        return self.inference(input_tensor)

    def inference(self, image_tensor):
        """
        image_tensor = image.transpose(2, 0, 1)
        image_tensor = image_tensor[np.newaxis, :]
        :param image_tensor:
        :return:
        """
        input_tensor = [image_tensor]
        outputs = self.forward(input_tensor)
        return outputs

    def forward(self, input_tensor):
        device_context.push()
        # Allocate device memory for inputs. This can be easily re-used if the
        # input shapes don't change  为输入变量赋予host空间，该空间可复用
        device_inputs = [cuda.mem_alloc(h_input.nbytes) for h_input in input_tensor]
        # Copy host inputs to device, this needs to be done for each new input， 由host拷贝到device
        for h_input, d_input in zip(input_tensor, device_inputs):
            cuda.memcpy_htod(d_input, h_input)
        # This needs to be called everytime your input shapes change
        # If your inputs are always the same shape (same batch size, etc.),
        # then you will only need to call this once 重新指定网络输入输出的大小。
        host_outputs, device_outputs = self.setup_binding_shapes(self.engine, self.context, input_tensor,
                                                                 self.input_binding_idxs,
                                                                 self.output_binding_idxs)  # 返回的是输出的idx和device buffer

        # print("\tOutput shapes: {}".format([out.shape for out in host_outputs]))
        # Bindings are a list of device pointers for inputs and outputs
        bindings = device_inputs + device_outputs  # list的合并
        # Inference
        self.context.execute_v2(bindings)
        # Copy outputs back to host to view results 将输出由gpu拷贝到cpu。
        for h_output, d_output in zip(host_outputs, device_outputs):
            cuda.memcpy_dtoh(h_output, d_output)
        # 释放显存
        for b in bindings: b.free()
        device_context.pop()
        return host_outputs

    @staticmethod
    def get_binding_idxs(engine: trt.ICudaEngine, profile_index: int):
        # Calculate start/end binding indices for current context's profile
        num_bindings_per_profile = engine.num_bindings // engine.num_optimization_profiles
        start_binding = profile_index * num_bindings_per_profile
        end_binding = start_binding + num_bindings_per_profile
        print("Engine/Binding Metadata")
        print("\tNumber of optimization profiles: {}".format(engine.num_optimization_profiles))
        print("\tNumber of bindings per profile : {}".format(num_bindings_per_profile))
        print("\tFirst binding for profile {}   : {}".format(profile_index, start_binding))
        print("\tLast binding for profile {}    : {}".format(profile_index, end_binding - 1))

        # Separate input and output binding indices for convenience
        input_binding_idxs = []
        output_binding_idxs = []
        for binding_index in range(start_binding, end_binding):
            if engine.binding_is_input(binding_index):
                input_binding_idxs.append(binding_index)
            else:
                output_binding_idxs.append(binding_index)

        return input_binding_idxs, output_binding_idxs

    @staticmethod
    def setup_binding_shapes(
            engine: trt.ICudaEngine,
            context: trt.IExecutionContext,
            host_inputs: List[np.ndarray],
            input_binding_idxs: List[int],
            output_binding_idxs: List[int],
    ):
        # 指定输入的shape，同时根据输入的shape指定输出的shape，并未输出赋予cuda空间
        # Explicitly set the dynamic input shapes, so the dynamic output
        # shapes can be computed internally
        for host_input, binding_index in zip(host_inputs, input_binding_idxs):
            context.set_binding_shape(binding_index, host_input.shape)
        # assert context.all_binding_shapes_specified
        host_outputs = []
        device_outputs = []
        for binding_index in output_binding_idxs:
            output_shape = context.get_binding_shape(binding_index)
            # Allocate buffers to hold output results after copying back to host
            buffer = np.empty(output_shape, dtype=np.float32)
            host_outputs.append(buffer)
            # Allocate output buffers on device
            device_outputs.append(cuda.mem_alloc(buffer.nbytes))
        return host_outputs, device_outputs


def onnx2trt(onnx_file, trt_file=None, input_shape=(1, 3, 160, 160), max_batch_size=8, quant=1, calibrator=None):
    """
    :param onnx_file: ONNX模型文件
    :param trt_file: 输出TRT模型文件，默认为onnx_file同目录
    :param input_shape: 模型输入维度
    :param quant: 0:不进行量化，1:进行半精度量化，2:进行INT8量化
    :param calibrator: 校准模块，INT8量化时需要
    :return:
    """

    def GiB(val):
        return val * 1 << 30

    if not trt_file: trt_file = onnx_file[:-len("onnx")] + "trt"
    batch_size, C, H, W = input_shape
    explicit_batch = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    with trt.Builder(TRT_LOGGER) as builder, \
            builder.create_network(explicit_batch) as network, \
            trt.OnnxParser(network, TRT_LOGGER) as parser:
        builder.max_batch_size = max_batch_size  # trt推理时最大支持的batchsize
        config = builder.create_builder_config()
        config.max_workspace_size = GiB(4)
        if quant == 1:
            # assert (builder.platform_has_fast_fp16 == True), 'not support fp16'
            config.set_flag(trt.BuilderFlag.FP16)
        elif quant == 2:
            # assert (builder.platform_has_fast_int8 == True), 'not support int8'
            config.set_flag(trt.BuilderFlag.INT8)
            config.int8_calibrator = calibrator
        print('Loading ONNX file from path {}...'.format(onnx_file))
        with open(onnx_file, 'rb') as model:
            print('Beginning ONNX file parsing')
            parser.parse(model.read())
        print('Completed parsing of ONNX file')
        print('Building an engine from file {}; this may take a while...'.format(onnx_file))
        profile = builder.create_optimization_profile()  # 动态输入时候需要 分别为最小输入、常规输入、最大输入
        # 有几个输入就要写几个profile.set_shape 名字和转onnx的时候要对应
        # tensorrt6以后的版本是支持动态输入的，需要给每个动态输入绑定一个profile，
        # 用于指定最小值常规值和最大值，如果超出这个范围会报异常。
        # min=(1, C, H, W), opt=(batch_size, C, H, W),max=(batch_size, C, H, W)
        inputs = [network.get_input(i) for i in range(network.num_inputs)]
        for inp in inputs:
            profile.set_shape(inp.name, (1, C, H, W), (batch_size, C, H, W), (batch_size, C, H, W))
        config.add_optimization_profile(profile)
        config.set_calibration_profile(profile)
        engine = builder.build_engine(network, config)
    print("Completed creating Engine")
    # 保存engine文件
    with open(trt_file, "wb") as f:
        f.write(engine.serialize())
    print("save TRT file:{}".format(trt_file))
    return engine


def test_performance(image_dir="./test.png", itera=10000):
    from pybaseutils import file_utils, time_utils
    input_size = (518, 518)
    # model_file = "./yolov5s_640_640.trt"
    model_file = "./yolov5s_640_640_fp16.trt"
    # model_file = "./yolov5s_640_640_int8.trt"
    input_shape = (1, 3, input_size[1], input_size[0])  # (B,C,H,W)
    engine = TRTEngine(model_file, input_shape=input_shape, quant=2, calibrator=None)
    image_list = file_utils.get_files_lists(file_dir=image_dir)
    image_list = image_list * itera
    for i in range(itera):
        image = cv2.imread(image_list[i])
        image = cv2.resize(image, (input_size[1], input_size[0]))
        image = image.transpose((2, 0, 1)).astype(np.float32) / 255.0
        inp_tensor = np.array([image], dtype=np.float32)
        with time_utils.Performance() as p:
            out_tensor = engine(inp_tensor)[0]
        print("inp_tensor:{}".format(inp_tensor.shape))
        print("out_tensor:{}".format(out_tensor.shape))


def test_example():
    from basetrainer.engine.trt_calibrator import ImageDataset, Calibrator, build_transform
    input_size = (640, 640)
    image_dir = '/home/PKing/nasdata/tmp/tmp/face_person/VOC/VOC2012/JPEGImages'
    dataset = ImageDataset(image_dir=image_dir, transform=build_transform, input_size=input_size)
    calibrator = Calibrator(dataset, cache_file="./calibration.cache")

    model_file = "./yolov5s_640_640.onnx"
    input_shape = (1, 3, input_size[1], input_size[0])
    engine = TRTEngine(model_file, input_shape=input_shape, quant=2, calibrator=calibrator)
    inp_tensor = np.ones(shape=input_shape, dtype=np.float32)
    out_tensor = engine(inp_tensor)[0]
    print("inp_tensor:{}".format(inp_tensor.shape))
    print("out_tensor:{}".format(out_tensor.shape))


if __name__ == "__main__":
    test_performance()
    # test_example()



import numpy as np
import torch


def area_of(left_top, right_bottom) -> torch.Tensor:
    """Compute the areas of rectangles given two corners.

    Args:
        left_top (N, 2): left top corner.
        right_bottom (N, 2): right bottom corner.

    Returns:
        area (N): return the area.
    """
    hw = torch.clamp(right_bottom - left_top, min=0.0)
    return hw[..., 0] * hw[..., 1]


def iou_of(boxes0, boxes1, eps=1e-5):
    """Return intersection-over-union (Jaccard index) of boxes.

    Args:
        boxes0 (N, 4): ground truth boxes.
        boxes1 (N or 1, 4): predicted boxes.
        eps: a small number to avoid 0 as denominator.
    Returns:
        iou (N): IoU values.
    """
    if isinstance(boxes0, np.ndarray):
        boxes0 = torch.from_numpy(boxes0)
    if isinstance(boxes1, np.ndarray):
        boxes1 = torch.from_numpy(boxes1)
    overlap_left_top = torch.max(boxes0[..., :2], boxes1[..., :2])
    overlap_right_bottom = torch.min(boxes0[..., 2:], boxes1[..., 2:])

    overlap_area = area_of(overlap_left_top, overlap_right_bottom)
    area0 = area_of(boxes0[..., :2], boxes0[..., 2:])
    area1 = area_of(boxes1[..., :2], boxes1[..., 2:])
    return overlap_area / (area0 + area1 - overlap_area + eps)


def compute_average_precision(precision, recall):
    """
    It computes average precision based on the definition of Pascal Competition. It computes the under curve area
    of precision and recall. Recall follows the normal definition. Precision is a variant.
    pascal_precision[i] = typical_precision[i:].max()
    """
    # identical but faster version of new_precision[i] = old_precision[i:].max()
    precision = np.concatenate([[0.0], precision, [0.0]])
    for i in range(len(precision) - 1, 0, -1):
        precision[i - 1] = np.maximum(precision[i - 1], precision[i])

    # find the index where the value changes
    recall = np.concatenate([[0.0], recall, [1.0]])
    changing_points = np.where(recall[1:] != recall[:-1])[0]

    # compute under curve area
    areas = (recall[changing_points + 1] - recall[changing_points]) * precision[changing_points + 1]
    return areas.sum()


def compute_voc2007_average_precision(precision, recall):
    ap = 0.
    for t in np.arange(0., 1.1, 0.1):
        if np.sum(recall >= t) == 0:
            p = 0
        else:
            p = np.max(precision[recall >= t])
        ap = ap + p / 11.
    return ap


def compute_average_precision_per_class(gt_boxes,
                                        pred_result: dict,
                                        iou_threshold,
                                        use_2007_metric,
                                        name="",
                                        plot=False):
    """
    :param num_true_cases:
    :param pred_result:
    :param iou_threshold:
    :param use_2007_metric:
    :param name:
    :param plot:
    :return:
    """
    num_true_cases = [len(gt) for gt in gt_boxes.values()]
    num_true_cases = sum(num_true_cases)
    image_ids = []
    boxes = []
    scores = []
    for file, item in pred_result.items():
        for line in item:
            image_ids.append(file)
            scores.append(line[0])
            boxes.append(line[1:])
    if len(scores) == 0:
        result = {"ap": 0}
        return result
    scores = np.asarray(scores, dtype=np.float32)
    sorted_indexes = np.argsort(-scores)
    boxes = [boxes[i] for i in sorted_indexes]
    image_ids = [image_ids[i] for i in sorted_indexes]
    true_positive = np.zeros(len(image_ids), dtype=np.float32)
    false_positive = np.zeros(len(image_ids), dtype=np.float32)
    matched = set()
    for i, image_id in enumerate(image_ids):
        box = np.asarray(boxes[i])
        if image_id not in gt_boxes:
            false_positive[i] = 1
            continue

        gt_box = np.asarray(gt_boxes[image_id])
        ious = iou_of(box, gt_box)
        max_iou = torch.max(ious).item()
        max_arg = torch.argmax(ious).item()
        if max_iou > iou_threshold:
            if (image_id, max_arg) not in matched:
                true_positive[i] = 1
                matched.add((image_id, max_arg))
            else:
                false_positive[i] = 1
        else:
            false_positive[i] = 1

    true_positive = true_positive.cumsum()
    false_positive = false_positive.cumsum()
    precision = true_positive / (true_positive + false_positive)
    recall = true_positive / num_true_cases

    if use_2007_metric:
        ap = compute_voc2007_average_precision(precision, recall)
    else:
        ap = compute_average_precision(precision, recall)
    sorted_scores = scores[sorted_indexes]
    result = {"ap": ap,
              "recall": recall,
              "precision": precision,
              "sorted_scores": sorted_scores,
              "true_positive": true_positive,
              "false_positive": false_positive,
              "num_true_cases": num_true_cases
              }

    if plot:
        plot_ap_recall_precision(ap, recall, precision, sorted_scores, name)
    # print_info(result, name)
    return result


def print_info(result, name, target_scores=0.5):
    """
    :param result:
    :param name:
    :param target_scores:
    :return:
    """
    info = "--" * 20
    ap = result["ap"]
    sorted_scores = result["sorted_scores"]
    true_positive = result["true_positive"]
    false_positive = result["false_positive"]
    num_true_cases = result["num_true_cases"]
    recall = result["recall"]
    precision = result["precision"]
    index = get_nearest_neighbor(sorted_scores, target=target_scores, k=1)
    tp_fp = true_positive + false_positive
    info += "\nscores:{}".format(sorted_scores[index])
    info += "\nground true:{}".format(num_true_cases)
    info += "\ntrue_positive:{}".format(true_positive[index])
    info += "\nfalse_positive:{}".format(false_positive[index])
    info += "\ntp_fp:{}".format(tp_fp[index])

    info += "\nrecall:{}".format(recall[index])
    info += "\nprecision:{}".format(precision[index])
    info += "\n{} AP:{}\n".format(name, ap)
    info += "--" * 20
    print(info)
    return info


def get_nearest_neighbor(data, target, k=1):
    """
    :param data:
    :param target:
    :param k:
    :return:
    """
    assert k == 1
    if not isinstance(data, np.ndarray):
        data = np.asarray(data)
    diff = np.abs(data - target)
    index = np.argmin(diff)
    return index


def plot_map_recall_precision(map_result: dict):
    """
    :param map_result:
    :return:
    """
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('TkAgg')
    class_name = list(map_result.keys())

    # Precision-Recall
    plt.figure()
    aps = []
    for name in class_name:
        recall = map_result[name]["recall"]
        precision = map_result[name]["precision"]
        ap = map_result[name]["ap"]
        recall = [0] + recall.tolist() + [1]
        precision = [1] + precision.tolist() + [0]
        plt.plot(recall, precision, label='{} AP:{:3.5f}%'.format(name, ap * 100))
        aps.append(ap)
    map = sum(aps) / len(aps)
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.title('Precision-Recall curve\nmAP={:3.5f}'.format(map))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    # Recall-Scores
    plt.figure()
    for name in class_name:
        recall = map_result[name]["recall"]
        sorted_scores = map_result[name]["sorted_scores"]
        recall = [0] + recall.tolist() + [1]
        sorted_scores = [1] + sorted_scores.tolist() + [0]
        plt.plot(sorted_scores, recall, label='{} Recall'.format(name))
    plt.xlabel('Scores')
    plt.ylabel('Recall')
    plt.title('Recall-Scores')
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()
    # Precision-Scores
    plt.figure()
    for name in class_name:
        precision = map_result[name]["precision"]
        sorted_scores = map_result[name]["sorted_scores"]
        precision = [1] + precision.tolist() + [0]
        sorted_scores = [1] + sorted_scores.tolist() + [0]
        plt.plot(sorted_scores, precision, label='{} Precision'.format(name))
    plt.xlabel('Scores')
    plt.ylabel('Precision')
    plt.title('Precision-Scores')
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()
    plt.show()


def plot_ap_recall_precision(ap, recall, precision, sorted_scores, name):
    """
    sorted_scores = scores[sorted_indexes]
    :param ap:
    :param recall:
    :param precision:
    :param sorted_scores:
    :param name:
    :return:
    """
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib.use('TkAgg')

    # 逆序排序
    recall = [0] + recall.tolist() + [1]
    precision = [1] + precision.tolist() + [0]
    sorted_scores = [1] + sorted_scores.tolist() + [0]
    plt.figure()
    plt.plot(recall, precision, label='Precision')
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    ap_str = "{0:.2f}%".format(ap * 100)
    # ap_str = "{0:.4f}%".format(ap * 100)
    plt.title('Precision x Recall curve \nClass: {}, AP: {:3.5f}'.format(name, ap_str))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.figure()
    plt.plot(sorted_scores, recall, label='Recall')
    plt.xlabel('Scores')
    plt.ylabel('Recall')
    plt.title('Recall-Scores\nClass:{}'.format(name))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.figure()
    plt.plot(sorted_scores, precision, label='Precision')
    plt.xlabel('Scores')
    plt.ylabel('Precision')
    plt.title('Precision-Scores\nClass:{}'.format(name))
    # plt.title('Precision x Recall curve \nClass: %s' % str(classId))
    plt.legend(shadow=True)
    plt.xlim([0, 1.05])
    plt.ylim([0, 1.05])
    plt.grid()

    plt.show()
    # plt.waitforbuttonpress()
    plt.pause(0.05)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-08-12 20:31:19
"""
from .MultiStepLR import MultiStepLR
from .CosineAnnealingLR import CosineAnnealingLR
from .ExponentialLR import ExponentialLR
from .LambdaLR import LambdaLR


def get_scheduler(scheduler, optimizer, lr_init, num_epochs, num_steps, **kwargs):
    if scheduler.lower() == "multi-step".lower() or scheduler.lower() == "multi_step".lower():
        lr_scheduler = MultiStepLR(optimizer,
                                   lr_init=lr_init,
                                   epochs=num_epochs,
                                   num_steps=num_steps,
                                   milestones=kwargs["milestones"],
                                   num_warn_up=kwargs["num_warn_up"])
    elif scheduler.lower() == "cosine".lower():
        # 余弦退火学习率调整策略
        lr_scheduler = CosineAnnealingLR(optimizer,
                                         num_epochs,
                                         num_steps=num_steps,
                                         lr_init=lr_init,
                                         num_warn_up=kwargs["num_warn_up"],
                                         num_cycles=kwargs.get("num_cycles", 1),
                                         decay=kwargs.get("decay", 1.0),
                                         )
    elif scheduler.lower() == "ExpLR".lower() or scheduler.lower() == "exp".lower():
        # 指数衰减学习率
        lr_scheduler = ExponentialLR(optimizer,
                                     num_epochs,
                                     num_steps=num_steps,
                                     lr_init=lr_init,
                                     num_warn_up=kwargs["num_warn_up"],
                                     decay=kwargs.get("decay", 0.95),
                                     )
    elif scheduler.lower() == "LambdaLR".lower():
        # 指数衰减学习率
        lr_scheduler = LambdaLR(optimizer,
                                num_epochs,
                                num_steps=num_steps,
                                lr_init=lr_init,
                                num_warn_up=kwargs["num_warn_up"]
                                )
    else:
        raise Exception("Error: scheduler type: {}".format(scheduler))
    return lr_scheduler



# -*-coding: utf-8 -*-
"""
    @Author : Pan
    @E-mail :
    @Date   : 2021-04-21 09:41:34
"""
import random
import numpy as np
import math


class ResampleExample(object):
    """样本均衡，重采样DataResampler的使用方法"""

    def __init__(self, item_list, label_index=1, shuffle=True, disp=False):
        """
        :param item_list: item_list=[item_0,item_1,...,item_n],
                          item_n= [path/to/image,label]
        :param label_index: label在item_n的index
        :param disp: 是否打印log信息
        """
        self.disp = disp
        self.shuffle = shuffle
        self.item_list = item_list
        self.resampler = DataResample(self.item_list,
                                      label_index=label_index,
                                      shuffle=self.shuffle,
                                      disp=self.disp)

    def __len__(self):
        # 更新resampler，实现每个epoch重新采样，避免样本数比较多的类别，没有加入训练
        self.item_list = self.resampler.update(self.shuffle)
        return len(self.item_list)

    def __getitem__(self, idx):
        image_path = self.item_list[idx][0]
        label_id = self.item_list[idx][1]
        return image_path, label_id


class DataResample(object):
    """样本均衡，重采样的方法"""

    def __init__(self, item_list=[], label_index=1, balance="mean", shuffle=True, disp=False):
        """
        Usage:
        参考：ResampleExample例子的使用方法
        :param item_list:
        :param label_index:
        :param balance:实现样本均衡策略,均衡力度：mean > log > sqrt > y
                        "y": 每个label样本数跟原来一样
                        "sqrt": 每个label样本取sqrt数，实现样本均衡
                        "log": 每个label样本取log数，实现样本均衡
                        "mean": 每个label样本取样本平均数，每个label的个数一样
        """
        self.src_item_list = item_list
        self.label_index = label_index
        self.balance = balance
        self.shuffle = shuffle
        self.disp = disp
        self.class_count = self.get_class_count(self.src_item_list, label_index)
        self.class_item_dict = self.get_class_item_dict(self.src_item_list, label_index)
        self.balance_nums = self.get_balance_nums(self.class_count, self.balance)
        self.item_list = self.update(shuffle=self.shuffle)
        self.class_weight = self.get_class_weight(self.class_count)

    def __len__(self):
        self.update(shuffle=self.shuffle)
        return len(self.item_list)

    def update(self, shuffle=False):
        self.item_list = self.get_resample_data(shuffle=shuffle)
        return self.item_list

    def get_resample_data(self, shuffle=True):
        """
        获得重采样的数据
        :param item_list:
        :param label_index:
        :param shuffle:
        :return:
        """
        if self.disp:
            print("class_item_dict:{}".format({k: len(v) for k, v in self.class_item_dict.items()}))
        out_list = []
        for name, per_class_list in self.class_item_dict.items():
            nums = self.balance_nums[name]
            per_list = self.get_sampler(per_class_list, nums, shuffle=shuffle)
            out_list += per_list
        if shuffle:
            random.shuffle(out_list)
        if self.disp:
            # 统计每个类别的个数
            class_count = self.get_class_count(out_list, self.label_index)
            print("resampler count_class :{},total:{}".format(class_count, sum(class_count.values())))
        return out_list

    def get_balance_nums(self, class_count: dict, balance):
        """
        获得平衡后，每个样本的数目
        :param class_count:
        :param balance:
        :return:
        """
        class_name = list(class_count.keys())
        num_samples = sum(class_count.values())  # 总样本数目
        if balance == "mean":
            mean_samples = num_samples * 1.0 / len(class_name)  # 平均样本数
            balance_nums = {name: mean_samples for name, c in class_count.items()}
        elif balance == "log":
            # Fix Bug:c=1
            balance_nums = {name: np.log(c + 1) for name, c in class_count.items()}
        elif balance == "sqrt":
            # Fix Bug:c=0
            balance_nums = {name: np.sqrt(c + 1) for name, c in class_count.items()}
        elif balance == "y":
            balance_nums = {name: c for name, c in class_count.items()}
        else:
            raise Exception("Error:{}".format(balance))
        sum_balance = sum(balance_nums.values())
        balance_nums = {name: math.ceil(c / sum_balance * num_samples) for name, c in balance_nums.items()}
        return balance_nums

    def get_sampler(self, item_list, nums, shuffle=True):
        """
        提取nums个数，不足nums个时，会进行填充
        :param item_list: 输入样本列表
        :param nums: 需要提取的样本数目
        :param shuffle: 是否随机提取样本
        :return:
        """
        item_nums = len(item_list)
        if nums > item_nums:
            item_list = item_list * math.ceil(nums / item_nums)
        if shuffle:
            random.shuffle(item_list)
        out_list = item_list[:nums]
        return out_list

    @staticmethod
    def get_label_list(item_list, label_index):
        labels_list = []
        for item in item_list:
            label = item[label_index]
            labels_list.append(label)
        return labels_list

    @staticmethod
    def get_class_item_dict(item_list, label_index):
        """
        获得每一类的样本
        :param item_list:
        :return:
        """
        class_item_dict = {}
        for item in item_list:
            label = item[label_index]
            try:
                # if label in class_item_dict: # 比较慢，相当于需要查询label是否存在
                class_item_dict[label] += [item]
            except Exception as e:
                class_item_dict[label] = [item]
        return class_item_dict

    @staticmethod
    def get_class_count(item_list, label_index):
        """
        统计每个类别的个数
        :param item_list:
        :param label_index: label在item中的序号
        :return:
        """
        class_count = {}
        for item in item_list:
            label = item[label_index]
            try:
                # if label in class_count:  # 比较慢，相当于需要查询label是否存在
                class_count[label] += 1
            except Exception as e:
                class_count[label] = 1
        return class_count

    @staticmethod
    def get_class_weight(class_count: dict):
        """
        计算每个label的权重，类别越少，权重越大
        :param class_count:
        :return:
        """
        n_samples = sum(list(class_count.values()))
        class_weight = {}
        for cls, num in class_count.items():
            score = n_samples / num
            class_weight[cls] = score
        return class_weight


if __name__ == "__main__":
    from torch.utils.data import Dataset, DataLoader, Sampler
    from utils import torch_tools

    torch_tools.set_env_random_seed()
    label0 = [["0.1.jpg", 0], ["0.2.jpg", 0], ["0.3.jpg", 0]]
    label1 = [["1.jpg", 1]] * 5
    label2 = [["2.0.jpg", 2], ["2.1.jpg", 2], ["2.2.jpg", 2], ["2.3.jpg", 2], ["2.4.jpg", 2],
              ["2.5.jpg", 2], ["2.6.jpg", 2], ["2.7.jpg", 2], ["2.8.jpg", 2]]
    label3 = [["3.1.jpg", 3], ["3.2.jpg", 3], ["3.3.jpg", 3]]
    item_list = label0 + label1 + label2 + label3
    item_list = item_list * 1000000
    print("have item_list:{}".format(len(item_list)))
    dataset_train = ResampleExample(item_list=item_list, label_index=1, shuffle=True, disp=True)
    # dataset_train = ResampleExample(item_list=item_list, label_index=1, disp=False)
    # dataset_train = ResampleExample(item_list=item_list, label_index=1, disp=False)
    batch_size = 4
    dataloader = DataLoader(dataset_train, batch_size, num_workers=0)
    epochs = 2
    for epoch in range(epochs):
        print("{}===".format(epoch) * 10)
        for batch_image, batch_label in iter(dataloader):
            print(batch_image, batch_label)
